/*
 *  CRC16.cpp - CRC 16 checksum class
 *
 *  Copyright (C) 2003 Marcel Mueller
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "CRC16.h"

#include <assert.h>


void CRC16::Update8(unsigned char value, size_t bits)
{  assert(bits-1 < 8); // matches 1..8 (unsigned compare!)
   value <<= 9;
   CRC <<= 1;
   switch (bits)
   {default:
      if ((CRC ^ value) & 0x10000)
         CRC ^= Polynomial;
      value <<= 1;
      CRC <<= 1;
    case 7:
      if ((CRC ^ value) & 0x10000)
         CRC ^= Polynomial;
      value <<= 1;
      CRC <<= 1;
    case 6:
      if ((CRC ^ value) & 0x10000)
         CRC ^= Polynomial;
      value <<= 1;
      CRC <<= 1;
    case 5:
      if ((CRC ^ value) & 0x10000)
         CRC ^= Polynomial;
      value <<= 1;
      CRC <<= 1;
    case 4:
      if ((CRC ^ value) & 0x10000)
         CRC ^= Polynomial;
      value <<= 1;
      CRC <<= 1;
    case 3:
      if ((CRC ^ value) & 0x10000)
         CRC ^= Polynomial;
      value <<= 1;
      CRC <<= 1;
    case 2:
      if ((CRC ^ value) & 0x10000)
         CRC ^= Polynomial;
      value <<= 1;
      CRC <<= 1;
    case 1:
      if ((CRC ^ value) & 0x10000)
         CRC ^= Polynomial;
   }
}

void CRC16::Update(unsigned char* data, size_t bits)
{  while (bits > 8)
   {  Update8(*data++);
      bits -= 8;
   }
   Update8(*data, bits);
}

