/*
 *  mp3gain.c - analyzes mp3 files, determines the perceived volume,
 *              and adjusts the volume of the mp3 accordingly
 *
 *  Copyright (C) 2002-2004 Glen Sawyer & Marcel Mueller
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  coding by Glen Sawyer (glensawyer@hotmail.com) 442 N 700 E, Provo, UT 84606 USA
 *    -- go ahead and laugh at me for my lousy coding skillz, I can take it :)
 *       Just do me a favor and let me know how you improve the code.
 *       Thanks.

 *  Ported to ANSI C (or at least somethingvery similar) by Marcel M�ller.
 *  /p won't work in this port, because there is no standard way to overwrite
 *  the last modified time stamp.
 */


/*
 *  General warning: I coded this in several stages over the course of several
 *  months. During that time, I changed my mind about my coding style and
 *  naming conventions many, many times. So there's not a lot of consistency
 *  in the code. Sorry about that. I may clean it up some day, but by the time
 *  I would be getting around to it, I'm sure that the more clever programmers
 *  out there will have come up with superior versions already...
 *
 *  So have fun dissecting.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <list>

#if defined(WIN32) || defined(__OS2__)
#define SWITCH_CHAR '/' // '-' is always accepted as switch char to keep some compatibility.
#endif

#include "interface.h"

#include "mp3gain.h"  /*jzitt*/
#include "rg_error.h" /*jzitt*/

#include "filechange.h"
#include "filefuncs.h"


static const char* progname;


// command line stuff

static void showVersion()
{  fprintf(stderr,"%s version %s\n",progname,MP3GAIN_VERSION);
}

static void errUsage()
{  showVersion();
   fprintf(stderr,"copyright(c) 2002-2004 by Glen Sawyer, Marcel M�ller\n"
                  "uses mpglib, which can be found at http://www.mpg123.de\n\n"
                  "Usage: %s [options] <infile> [<infile 2> ...]\n\n"
                  "  --use \"%s -?\" for a full list of options\n",progname,progname);
   fflush(stderr);
   exit(48);
}

static void fullUsage()
{  showVersion();
   fprintf(stderr,"copyright(c) 2002-2004 by Glen Sawyer, Marcel M�ller \n"
                  "uses mpglib, which can be found at http://www.mpg123.de\n\n"
                  "Usage: %s [options] <infile> [<infile 2> ...]\n\n"
                  "options:\n"
                  " -v - show version number\n"
                  " -g <i> - apply gain i to mp3 without doing any analysis\n"
                  "          Keep in mind that integer gain changes work differently on layer I\n"
                  "          and layer II audio files than with layer III.\n"
                  "          (use \"-? l2\" switch for more details)\n"
                  " -l 0 <i> - apply gain i to channel 0 (left channel) of mpeg without doing\n"
                  "            any analysis (ONLY works for STEREO files, not Joint Stereo)\n"
                  " -l 1 <i> - apply gain i to channel 1 (right channel) of mpeg\n\n"
                  " -r - apply Track gain automatically (all files set to equal loudness)\n"
                  " -a - apply Album gain automatically (files are all from the same album: a\n"
                  "      single gain change is applied to all files, so their loudness relative\n"
                  "      to each other remains unchanged, but the average album loudness is\n"
                  "      normalized)\n"
   //             " -m <i> - modify suggested MP3 gain by integer i\n"
                  " -d <n> - modify suggested dB gain by floating-point n\n"
                  " -k - automatically lower track gain to not clip audio\n"
                  "      By default the gain is only not increased if that will cause clipping.\n"
                  " -c - ignore clipping warning when applying gain\n"
                  " -C <r> - ignore clipping warning, auto reply r\n"
                  "          /Ca will not increase the volume over the clipping limit.\n"
                  " -i - interactive mode: allways commit manually\n"
                  " -I <r> - interactive mode: default reply r\n"
                  " -o - output is a database-friendly tab-delimited list\n"
                  " -t - mp3gain writes modified mpeg to temp file, then deletes original\n"
                  "      instead of modifying bytes in original file\n"
                  " -q - Quiet mode: no status messages\n"
                  " -p - Preserve original file timestamp\n"
                  " -x - Only find max. amplitude of mp3\n"
                  " -f - Force mp3gain to assume input file is an MPEG Layer III file\n"
                  "      (i.e. don't check for mis-named Layer I or Layer II files)\n"
                  " -w - \"wrap\" gain change if gain+change > 255 or gain+change < 0\n"
                  "      (use \"-? wrap\" switch for a complete explanation)\n"
                  " -? - show this message\n\n"
                  "if you specify -r and -a, only the second one will work.\n"
                  "This also aplies to -i, -I, -c and -C.\n"
                  "If you do not specify -c, the program will stop and ask before applying gain\n"
                  "change to a file that might clip. In case of -i it will always stop.\n\n"
   #ifdef SWITCH_CHAR
                  "Note: You may use %c<option> instead of -<option> on Windows and OS/2 platform."
                   ,progname, SWITCH_CHAR);
   #else
                   ,progname);
   #endif
   fflush(stderr);
   exit(0);
}

void wrapExplanation()
{  showVersion();
   fprintf(stderr,"Wrap gain - Here's the problem:\n"
                  "The \"global gain\" field that mp3gain adjusts is an 8-bit unsigned integer, so\n"
                  "the possible values are 0 to 255.\n"
                  "\n"
                  "MOST mp3 files (in fact, ALL the mp3 files I've examined so far) don't go\n"
                  "over 230. So there's plenty of headroom on top-- you can increase the gain\n"
                  "by 37dB (multiplying the amplitude by 76) without a problem.\n"
                  "\n"
                  "The problem is at the bottom of the range. Some encoders create frames with\n"
                  "0 as the global gain for silent frames.\n"
                  "What happens when you _lower_ the global gain by 1?\n"
                  "Well, in the past, mp3gain always simply wrapped the result up to 255.\n"
                  "That way, if you lowered the gain by any amount and then raised it by the\n"
                  "same amount, the mp3 would always be _exactly_ the same.\n"
                  "\n"
                  "There are a few encoders out there, unfortunately, that create 0-gain frames\n"
                  "with other audio data in the frame.\n"
                  "As long as the global gain is 0, you'll never hear the data.\n"
                  "But if you lower the gain on such a file, the global gain is suddenly _huge_.\n"
                  "If you play this modified file, there might be a brief, very loud blip.\n"
                  "\n"
                  "So now the default behavior of mp3gain is to _not_ wrap gain changes.\n"
                  "In other words,\n"
                  "1) If the gain change would make a frame's global gain drop below 0,\n"
                  "   then the global gain is set to 0.\n"
                  "2) If the gain change would make a frame's global gain grow above 255,\n"
                  "   then the global gain is set to 255.\n"
                  "3) If a frame's global gain field is already 0, it is not changed, even if\n"
                  "   the gain change is a positive number\n"
                  "\n"
                  "To use the original \"wrapping\" behavior, use the \"-w\" switch.\n"
                  "\n"
                  "Layer I and layer II files are even more sensitive to that since they only provide\n"
                  "scale factors between 63 and 0. It is not recommended to use -w for that files at all.\n"
                  "(use \"-? l2\" switch for more details)\n"
                  "\n");
   fflush(stderr);
   exit(0);
}

void l2Details()
{  showVersion();
   fprintf(stderr,"Handling of layer I and layer II mpeg audio files (mp1 & mp2):\n"
                  "\n"
                  "Layer I and layer II files do not have global gain fields. Therefore the only\n"
                  "way to adjust their gain without reencoding is to adjust the scale factors.\n"
                  "\n"
                  "The scale factors are in the range of 63 to 0. They have a resolution of 2dB.\n"
                  "So any changes made to layer I and II files must be multiples of 2dB.\n"
                  "The total dynamic range is >120dB - sounds very much, but this is not true\n"
                  "since the scalfactors are frequency dependent. Usually you may vary the gain of\n"
                  "layer I & II files only by some few dB until some of the scale factors leave\n"
                  "their domain.\n"
                  "Future versions of mp3gain may take care of this and warn you before that will\n"
                  "happen.\n"
                  "\n");
   fflush(stderr);
   exit(0);
}

// coomand line
static void FatalMsg(int no, const char* fmt, ...)
{  va_list va;
   va_start(va, fmt);
   vfprintf(stderr, fmt, va);
   va_end(va);
   errUsage();
   exit(no);
}

static double FloatOpt(const char**& src, const char*& par)
{  double r;
   size_t n = 0;
   sscanf(par, "%lf%zn", &r, &n);
   if (n)
   {  // inline option
      par += n;
      return r;
   }
   // try next param
   if (*src == NULL)
      FatalMsg(42, "Option %c expects a floating point parameter.\n", par[-1]);
   sscanf(*src, "%lf%zn", &r, &n);
   if (n != strlen(*src))
      FatalMsg(42, "Option %c expects a floating point parameter, found %s.\n", par[-1], *src);
   ++src;
   return r;
}

static int IntOpt(const char**& src, const char*& par)
{  int r;
   size_t n = 0;
   sscanf(par, "%i%zn", &r, &n);
   if (n)
   {  // inline option
      par += n;
      return r;
   }
   // try next param
   if (*src == NULL)
      FatalMsg(42, "Option %c expects an integer parameter.\n", par[-1]);
   sscanf(*src, "%i%zn", &r, &n);
   if (n != strlen(*src))
      FatalMsg(42, "Option %c expects an integer parameter, found %s.\n", par[-1], *src);
   ++src;
   return r;
}
static int IntOpt(const char**& src, const char*& par, int minval, int maxval)
{  int r = IntOpt(src, par);
   if (r < minval || r > maxval)
      FatalMsg(42, "Parameter %i out of range [%i,%i].\n", r, minval, maxval);
   return r;
}

static const char* StringOpt(const char**& src, const char*& par, const char* def)
{  if (*par)
      return par;
   if (*src == NULL)
      return def;
   return *++src;
}


int main(int argc, const char **argv)
{  // setup global default event handler
   EventLog mylog(std::cerr);
   DefEH = &mylog;

   int applyRadio = 0;
   int applyAlbum = 0;
   bool directGain = false;
   int directGainVal[2] = {0,0};
   //int mp3GainMod = 0;
   double dBGainMod = 0;

   maxAmpOnly = 0;

   progname = *argv++;

   if (argc < 2)
      errUsage();

   // parse command line
   const char** src = argv;   // to parse the options

   std::list<filechange> files; // avoid our own memory management

   while (*src)
   {  const char* param = *src++;
      if ( param[0] == '-'
         #ifdef SWITCH_CHAR
          || param[0] == SWITCH_CHAR
         #endif
          )
      {  ++param;
         while (param && *param)
         {  switch(*param++)
            {case 'a':
               applyRadio = 0;
               applyAlbum = true;
               break;

             case 'C':
               filechange::defInput = param;
               param = NULL; // ate
               filechange::ignoreClipWarning = true;
               break;
             case 'c':
               filechange::ignoreClipWarning = true;
               filechange::defInput = "y";
               break;

             case 'd':
               dBGainMod = FloatOpt(src, param);
               //fprintf(stderr,"dBGainMod = %f\n",dBGainMod);
               break;

             case 'f':
               filechange::ForceLayer = 3;
               break;

             case 'g':
               directGain = true;
               directGainVal[0] = directGainVal[1] = IntOpt(src, param);
               // fprintf(stderr,"directGainVal = %d\n",directGainVal);
               break;

             case 'I':
               filechange::defInput = param;
               param = NULL; // ate
             case 'i':
               filechange::InteractiveMode = true;
               break;

             case 'k':
               filechange::autoNoClip = true;
               filechange::ignoreClipWarning = true;
               if (filechange::defInput[0] == 0) // do not overwrite previously set values
                  filechange::defInput = "a";
               break;

             case 'l':
               {  directGain = true;
                  int whichChannel = IntOpt(src, param) & 1; // avoid array overflows
                  // fprintf(stderr,"directGainVal = %d\n",directGainVal);
                  directGainVal[whichChannel] = IntOpt(src, param);
                  break;
               }
             case 'm':
               FatalMsg(44, "The option -m is no longer supported!\n It leads to unpredictable results unless the MPEG layer is known.\nUse -d <dBchange> instead.");
               //mp3GainMod = IntOpt(src, param);
               // fprintf(stderr,"mp3GainMod = %d\n",mp3GainMod);
               break;

             case 'o':
               filechange::databaseFormat = true;
               break;

             case 'p':
               filechange::saveTime = true;
               //fprintf(stderr,"Option p not supported on ANSI-C++.\n");
               break;

             case 'q':
               filechange::QuietMode = true;
               break;

             case 'r':
               applyRadio = true;
               applyAlbum = false;
               break;

             case 't':
               filechange::UsingTemp = true;
               break;

             case 'x':
               maxAmpOnly = true;
               break;

             case 'h':
             case '?':
               switch (toupper(StringOpt(src, param, "")[0]))
               {  // none of the cases below will ever return
                default:
                  fullUsage();
                case 'W':
                  wrapExplanation();
                case 'L':
                  l2Details();
               }

             case 'v':
               showVersion();
               exit(0);

             case 'w':
               filechange::wrapGain = true;
               break;

             default:
               FatalMsg(44, "I don't recognize option %s\n", param[-1]);
            }
         }
      } else
         // no option! => filename
         files.push_back(filechange(param));
   }

   if (files.size() == 0)
      FatalMsg(10, "Nothing to do.\n");

   if (fileinfo::databaseFormat)
      fprintf(stdout,"File\tMP3 gain\tdB gain\tMax Amplitude\n");

   for (std::list<filechange>::iterator i = files.begin(); i != files.end(); ++i)
   {  if (directGain)
         i->ChangeGain(directGainVal[0], directGainVal[1]);
      // analyze !!
       else if (i->Analyze() == 0)
      {  if (maxAmpOnly)
            i->dBchange = 0;
          else
         {  i->dBchange = GetTitleGain();
            if (i->dBchange == GAIN_NOT_ENOUGH_SAMPLES)
            {  fprintf(stderr, "Not enough samples in %s to do analysis\n\n", i->filename);
               continue;
            }
            i->dBchange += dBGainMod;
         }

         if (filechange::InteractiveMode || (!applyRadio && !applyAlbum))
            i->PrintGainInfo();
         if (applyRadio)
            i->ChangeGain(i->CommitGainChange());
      }
   }

   if (!applyRadio && !directGain)
   {  filechange fi(NULL); // generate virtual fileinfo
      // init dBchange
      if (maxAmpOnly)
         fi.dBchange = 0;
       else
      {  fi.dBchange = GetAlbumGain();
         if (fi.dBchange == GAIN_NOT_ENOUGH_SAMPLES)
            FatalMsg(24, "Not enough samples in the files to do analysis\n\n");
         fi.dBchange += dBGainMod;
      }
      // calculate maxsample and check MPEG layer
      fi.maxsample = 0;
      fi.layer = 0; // undefined
      for (std::list<filechange>::iterator i = files.begin(); i != files.end(); ++i)
         if (i->fileOK)
         {  if (i->maxsample > fi.maxsample)
               fi.maxsample = i->maxsample;
            if (fi.layer == 0)
               fi.layer = i->layer;
             else if ((fi.layer == 3) ^ (i->layer == 3)) // for now only layer 3 is different
               fi.layer = 100; // some bad value, must be a multiple of 4
         }
      fi.layer &= 3;

      if (filechange::InteractiveMode || !applyAlbum)
         fi.PrintGainInfo();

      if (applyAlbum)
      {  if (fi.layer == 0)
            // files have different MPEG layers !!!
            FatalMsg(24, "Can't apply album gain to a set of files encoded with different MPEG layers.\n\n");

         int intGainChange = fi.CommitGainChange();
         if (intGainChange != 0)
            for (std::list<filechange>::iterator i = files.begin(); i != files.end(); ++i)
               if (i->fileOK)
                  i->ChangeGain(intGainChange);
      }
   }
   // done
   return 0;
}

